use base64::Engine;

#[macro_export]
macro_rules! bvec {
    ($s:expr) => {
        Vec::<u8>::from_hex(include_str!($s)).unwrap()
    };
}

#[macro_export]
macro_rules! bvec64 {
    ($s:expr) => {
        Vec::<u8>::from_b64(include_str!($s)).unwrap()
    };
}

#[macro_export]
macro_rules! bvecs {
    ($s:expr) => {
        include_str!($s).split("\n").filter(|l| *l!="").map(|p| 
            Vec::<u8>::from_hex(p).unwrap()
        ).collect::<Vec<Vec<u8>>>()
    };
}

#[derive(Debug)]
pub enum PreludeError {
    WrongLength(usize),
    B64Err(base64::DecodeError),
    FHErr(hex::FromHexError),
    Utf8Err(std::str::Utf8Error),
    Error(&'static str),
}

pub type CPals<T> = std::result::Result<T, PreludeError>;

pub trait ByteBased: Sized {
    fn from_hex(h: &str) -> CPals<Self>;
    fn to_hex(&self) -> String;

    fn from_b64(b: &str) -> CPals<Self>;
    fn to_b64(&self) -> String;

    fn from_text(t: &str) -> CPals<Self>;
    fn to_text(&self) -> CPals<&str>;

    fn xor_with(&self, other: &Self) -> CPals<Self>;
    fn xor_repeating(&self, other: &Self) -> CPals<Self>;

    fn hamming(&self, other: &Self) -> CPals<u64>;

    fn to_fixed_len<const N: usize>(&self) -> CPals<[u8; N]>;
    fn to_variable_len(&self) -> Vec<u8>;
}

impl ByteBased for Vec<u8> {
    fn from_hex(h: &str) -> CPals<Self> {
        hex::decode(remove_ws(h)).map_err(PreludeError::FHErr)
    }

    fn to_hex(&self) -> String {
        hex::encode(self)
    }

    fn from_b64(b: &str) -> CPals<Self> {
        base64::engine::general_purpose::STANDARD.decode(remove_ws(b)).map_err(PreludeError::B64Err)
    }

    fn to_b64(&self) -> String {
        base64::engine::general_purpose::STANDARD.encode(self)
    }

    fn from_text(tx: &str) -> CPals<Self> {
        Ok(tx.as_bytes().to_vec())
    }

    fn to_text(&self) -> CPals<&str> {
        std::str::from_utf8(self).map_err(PreludeError::Utf8Err)
    }

    fn xor_with(&self, other: &Self) -> CPals<Self> {
        if other.len() != self.len() { return Err(PreludeError::Error("couldn't xor, different lengths")) }

        let mut vec2 = vec![0; self.len()];
        for i in 0..self.len() {
            vec2[i] = self[i] ^ other[i];
        };
        Ok(vec2)
    }

    fn xor_repeating(&self, other: &Self) -> CPals<Self> {
        if other.len() == 0 { return Err(PreludeError::Error("couldn't xor, key was empty")) }
        let mut vec2 = vec![0; self.len()];
        for i in 0..self.len() {
            vec2[i] = self[i] ^ other[i % other.len()];
        };
        Ok(vec2)
    }

    fn hamming(&self, other: &Self) -> CPals<u64> {
        if other.len() != self.len() { return Err(PreludeError::Error("couldn't ham, different lengths")) }

        let mut bdist: u64 = 0;
        for i in 0..self.len() {
            bdist += (self[i] ^ other[i]).count_ones() as u64;
        }
        Ok(bdist)
    }

    fn to_fixed_len<const N: usize>(&self) -> CPals<[u8; N]> {
        if self.len() != N {
            return Err(PreludeError::WrongLength(self.len()))
        }
        // avoid the owned requirement for [u8; N]
        Ok(*self.array_chunks().next().unwrap())
    }

    fn to_variable_len(&self) -> Vec<u8> {
        self.clone()
    }
}

impl<const N: usize> ByteBased for [u8; N] {
    fn from_hex(h: &str) -> CPals<Self> {
        let v = <Vec<u8>>::from_hex(h)?;
        v.to_fixed_len()
    }

    fn to_hex(&self) -> String {
        self.to_variable_len().to_hex()
    }

    fn from_b64(b: &str) -> CPals<Self> {
        let v = <Vec<u8>>::from_b64(b)?;
        v.to_fixed_len()
    }

    fn to_b64(&self) -> String {
        self.to_variable_len().to_b64()
    }

    fn from_text(t: &str) -> CPals<Self> {
        let v = <Vec<u8>>::from_text(t)?;
        v.to_fixed_len()
    }

    fn to_text(&self) -> CPals<&str> {
        std::str::from_utf8(self).map_err(PreludeError::Utf8Err)
    }

    fn xor_with(&self, other: &Self) -> CPals<Self> {
        let mut result = [0; N];
        for i in 0..N {
            result[i] = self[i] ^ other[i];
        }
        Ok(result)
    }

    fn xor_repeating(&self, other: &Self) -> CPals<Self> {
        self.xor_with(other) // they necessarily have the same length
    }

    fn hamming(&self, other: &Self) -> CPals<u64> {
        let mut count: u64 = 0;

        for i in 0..N {
            count += (self[i] ^ other[i]).count_ones() as u64;
        }

        Ok(count)
    }

    fn to_fixed_len<const N2: usize>(&self) -> CPals<[u8; N2]> {
        let slice: &[u8] = self;
        if let Ok(x) = <&[u8; N2]>::try_from(slice) {
            return Ok(*x)
        }
        return Err(PreludeError::WrongLength(self.len())) 
    }

    fn to_variable_len(&self) -> Vec<u8> {
        self.to_vec()
    }
}


pub fn remove_ws(s: &str) -> String {
    let mut new = String::with_capacity(s.len());
    for c in s.chars() {
        if !c.is_whitespace() { new.push(c) }
    }
    new
}